"""Replay buffer for deep q network"""

import numpy as np


class ReplayBuffer:
    """a circular queue based on numpy array, supporting batch put and batch get"""
    def __init__(self, shape, dtype=np.float32):
        self.buffer = np.empty(shape=shape, dtype=dtype)
        self.head   = 0
        self.capacity   = len(self.buffer)

    def put(self, data):
        """put data to

        Parameters
        ----------
        data: numpy array
            data to add
        """
        head = self.head
        n = len(data)
        if head + n <= self.capacity:
            self.buffer[head:head+n] = data
            self.head = (self.head + n) % self.capacity
        else:
            split = self.capacity - head
            self.buffer[head:] = data[:split]
            self.buffer[:n - split] = data[split:]
            self.head = split
        return n

    def get(self, index):
        """get items

        Parameters
        ----------
        index: int or numpy array
            it can be any numpy supported index
        """
        return self.buffer[index]

    def clear(self):
        """clear replay buffer"""
        self.head = 0



"""test baselines in battle against"""

from search import do_task

task = [
    {
        "name": "against",
        "type": "single-search",
        "prefix": "python examples/train_against.py --train --save_every 100 --n_round 500",
        "arg_name": "--alg",
        "arg_value": ["a2c", "drqn", "dqn"]
    }
]

for item in task:
    do_task(item)
    print("%s done" % item['name'])

print("battle-against all done")
print("plot curve: python scripts/plot_many.py against-rec.out num 0")




"""
A cooperation game, tigers must attack a same deer simultaneously to get reward
"""

import magent


def get_config(map_size):
    gw = magent.gridworld
    cfg = gw.Config()

    cfg.set({"map_width": map_size, "map_height": map_size})
    cfg.set({"embedding_size": 10})

    deer = cfg.register_agent_type(
        "deer",
        {'width': 1, 'length': 1, 'hp': 5, 'speed': 1,
         'view_range': gw.CircleRange(1), 'attack_range': gw.CircleRange(0),
         'step_recover': 0.2,
         'kill_supply': 8,
         })

    tiger = cfg.register_agent_type(
        "tiger",
        {'width': 1, 'length': 1, 'hp': 10, 'speed': 1,
         'view_range': gw.CircleRange(4), 'attack_range': gw.CircleRange(1),
         'damage': 1, 'step_recover': -0.2,
         })

    deer_group  = cfg.add_group(deer)
    tiger_group = cfg.add_group(tiger)

    a = gw.AgentSymbol(tiger_group, index='any')
    b = gw.AgentSymbol(tiger_group, index='any')
    c = gw.AgentSymbol(deer_group,  index='any')

    # tigers get reward when they attack a deer simultaneously
    e1 = gw.Event(a, 'attack', c)
    e2 = gw.Event(b, 'attack', c)
    cfg.add_reward_rule(e1 & e2, receiver=[a, b], value=[1, 1])

    return cfg



"""deprecated"""

import ctypes
import numpy as np

from magent.model import BaseModel
from magent.c_lib import _LIB, as_int32_c_array, as_float_c_array


class RushPredator(BaseModel):
    def __init__(self, env, handle, attack_handle, *args, **kwargs):
        BaseModel.__init__(self, env, handle)

        self.attack_channel = env.get_channel(attack_handle)
        self.attack_base, self.view2attack = env.get_view2attack(handle)

        print("attack_channel", self.attack_channel)
        print("view2attack", self.view2attack)

    def infer_action(self, observations, *args, **kwargs):
        obs_buf = as_float_c_array(observations[0])
        hp_buf  = as_float_c_array(observations[1])
        n, height, width, n_channel = observations[0].shape
        buf = np.empty((n,), dtype=np.int32)
        act_buf = as_int32_c_array(buf)
        attack_channel = self.attack_channel
        attack_base = self.attack_base
        view2attack_buf = as_int32_c_array(self.view2attack)

        _LIB.rush_prey_infer_action(obs_buf, hp_buf, n, height, width, n_channel,
                                    act_buf, attack_channel, attack_base,
                                    view2attack_buf, ctypes.c_float(100.0))
        return buf



""" some utilities """

import math
import collections
import platform

import numpy as np
import logging
import collections
import os

from magent.builtin.rule_model import RandomActor


class EpisodesBufferEntry:
    """Entry for episode buffer"""
    def __init__(self):
        self.views = []
        self.features = []
        self.actions = []
        self.rewards = []
        self.terminal = False

    def append(self, view, feature, action, reward, alive):
        self.views.append(view.copy())
        self.features.append(feature.copy())
        self.actions.append(action)
        self.rewards.append(reward)
        if not alive:
            self.terminal = True


class EpisodesBuffer:
    """Replay buffer to store a whole episode for all agents
       one entry for one agent
    """
    def __init__(self, capacity):
        self.buffer = {}
        self.capacity = capacity
        self.is_full = False

    def record_step(self, ids, obs, acts, rewards, alives):
        """record transitions (s, a, r, terminal) in a step"""
        buffer = self.buffer
        index = np.random.permutation(len(ids))

        if self.is_full:  # extract loop invariant in else part
            for i in range(len(ids)):
                entry = buffer.get(ids[i])
                if entry is None:
                    continue
                entry.append(obs[0][i], obs[1][i], acts[i], rewards[i], alives[i])
        else:
            for i in range(len(ids)):
                i = index[i]
                entry = buffer.get(ids[i])
                if entry is None:
                    if self.is_full:
                        continue
                    else:
                        entry = EpisodesBufferEntry()
                        buffer[ids[i]] = entry
                        if len(buffer) >= self.capacity:
                            self.is_full = True

                entry.append(obs[0][i], obs[1][i], acts[i], rewards[i], alives[i])

    def reset(self):
        """ clear replay buffer """
        self.buffer = {}
        self.is_full = False

    def episodes(self):
        """ get episodes """
        return self.buffer.values()


# decay schedulers
def exponential_decay(now_step, total_step, final_value, rate):
    """exponential decay scheduler"""
    decay = math.exp(math.log(final_value)/total_step ** rate)
    return max(final_value, 1 * decay ** (now_step ** rate))


def linear_decay(now_step, total_step, final_value):
    """linear decay scheduler"""
    decay = (1 - final_value) / total_step
    return max(final_value, 1 - decay * now_step)


def piecewise_decay(now_step, anchor, anchor_value):
    """piecewise linear decay scheduler

    Parameters
    ---------
    now_step : int
        current step
    anchor : list of integer
        step anchor
    anchor_value: list of float
        value at corresponding anchor
    """
    i = 0
    while i < len(anchor) and now_step >= anchor[i]:
        i += 1

    if i == len(anchor):
        return anchor_value[-1]
    else:
        return anchor_value[i-1] + (now_step - anchor[i-1]) * \
                                   ((anchor_value[i] - anchor_value[i-1]) / (anchor[i] - anchor[i-1]))


# eval observation set generator
def sample_observation(env, handles, n_obs=-1, step=-1):
    """Sample observations by random actors.
    These samples can be used for evaluation

    Parameters
    ----------
    env : environment
    handles: list of handle
    n_obs : int
        number of observation
    step : int
        maximum step

    Returns
    -------
    ret : list of raw observation
        raw observation for every group
        the format of raw observation is tuple(view, feature)
    """
    models = [RandomActor(env, handle) for handle in handles]

    n = len(handles)
    views = [[] for _ in range(n)]
    features = [[] for _ in range(n)]

    done = False
    step_ct = 0
    while not done:
        obs = [env.get_observation(handle) for handle in handles]
        ids = [env.get_agent_id(handle) for handle in handles]

        for i in range(n):
            act = models[i].infer_action(obs[i], ids[i])
            env.set_action(handles[i], act)

        done = env.step()
        env.clear_dead()

        # record steps
        for i in range(n):
            views[i].append(obs[i][0])
            features[i].append(features[i][1])

        if step != -1 and step_ct > step:
            break

        if step_ct % 100 == 0:
            print("sample step %d" % step_ct)

        step_ct += 1

    for i in range(n):
        views[i] = np.array(views[i], dtype=np.float32).reshape((-1,) +
                            env.get_view_space(handles[i]))
        features[i] = np.array(features[i], dtype=np.float32).reshape((-1,) +
                               env.get_feature_space(handles[i]))

    if n_obs != -1:
        for i in range(n):
            views[i] = views[i][np.random.choice(np.arange(views[i].shape[0]), n_obs)]
            features[i] = features[i][np.random.choice(np.arange(features[i].shape[0]), n_obs)]

    ret = [(v, f) for v, f in zip(views, features)]
    return ret


def init_logger(filename):
    """ initialize logger config

    Parameters
    ----------
    filename : str
        filename of the log
    """
    logging.basicConfig(level=logging.INFO, filename=filename + ".log")
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    logging.getLogger('').addHandler(console)


def rec_round(x, ndigits=2):
    """ round x recursively

    Parameters
    ----------
    x: float, int, list, list of list, ...
        variable to round, support many types
    ndigits: int
        precision in decimal digits
    """
    if isinstance(x, collections.Iterable):
        return [rec_round(item, ndigits) for item in x]
    return round(x, ndigits)


def has_gpu():
    """ check where has a nvidia gpu """
    ret = os.popen("nvidia-smi -L 2>/dev/null").read()
    return ret.find("GPU") != -1


def download_file(filename, url):
    """download url to filename"""
    print("Download %s from %s..." % (filename, url))

    ret = os.system("wget -O %s '%s'" % (filename, url))

    if ret != 0:
        print("ERROR: wget fails!")
        print("If you are an OSX user, you can install wget by 'brew install wget' and retry.")
        exit(-1)
    else:
        print("download done!")


def download_model(url):
    """download model from url"""
    name = url.split('/')[-1]
    name = os.path.join('data', name)
    download_file(name, url)
    def do_commond(cmd):
        print(cmd)
        os.system(cmd)
    do_commond("tar xzf %s -C data" % name)
    do_commond("rm %s" % name)


def check_model(name):
    """check whether a model is downloaded"""
    infos = {
        'against':
            (('data/battle_model/battle/tfdqn_0.index',),
            'https://raw.githubusercontent.com/merrymercy/merrymercy.github.io/master/_data/magent/against-0.tar.gz'),

        'battle-game':
            (("data/battle_model/trusty-battle-game-l/tfdqn_0.index",
             "data/battle_model/trusty-battle-game-r/tfdqn_0.index"),
             'https://raw.githubusercontent.com/merrymercy/merrymercy.github.io/master/_data/magent/battle_model.tar.gz'),

        'arrange':
            (('data/arrange_model/arrange/tfdqn_10.index',),
             'https://raw.githubusercontent.com/merrymercy/merrymercy.github.io/master/_data/magent/arrange_game.tar.gz',)
    }

    if name not in infos:
        raise RuntimeError("Unknown model name")

    info = infos[name]
    missing = False
    for check in info[0]:
        if not os.path.exists(check):
            missing = True
    if missing:
        download_model(info[1])


class FontProvider:
    """provide pixel font"""
    def __init__(self, filename):
        data = []
        # read raw
        with open(filename) as fin:
            for line in fin.readlines():
                char = []
                for x in line.split(','):
                    char.append(eval(x))
                data.append(char)

        height = 8
        width  = 8

        # expand bit compress
        expand_data = []
        for char in data:
            expand_char = [[0 for _ in range(width)] for _ in range(height)]
            for i in range(width):
                for j in range(height):
                    set = char[i] & (1 << j)
                    if set:
                        expand_char[i][j] = 1
            expand_data.append(expand_char)

        self.data = expand_data
        self.width = width
        self.height = height

    def get(self, i):
        if isinstance(i, int):
            return self.data[i]
        else:
            return self.data[ord(i)]



from abc import ABCMeta, abstractmethod


class BaseServer:
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_info(self):
        pass

    @abstractmethod
    def get_data(self, frame_id, x_range, y_range):
        pass

    @abstractmethod
    def add_agents(self, x, y, g):
        pass
        
    @abstractmethod
    def get_map_size(self):
        pass

    @abstractmethod
    def get_banners(self, frame_id, resolution):
        pass

    @abstractmethod
    def get_status(self, frame_id):
        pass

    @abstractmethod
    def keydown(self, frame_id, key, mouse_x, mouse_y):
        pass

    @abstractmethod
    def mousedown(self, frame_id, key, mouse_x, mouse_y):
        pass

    @abstractmethod
    def get_endscreen(self, frame_id):
        pass


from .base_server import BaseServer


class SampleServer(BaseServer):
    def get_group_info(self):
        return [[1, 1, 0, 0, 0]]

    def get_static_info(self):
        return {"walls": []}

    def get_data(self, frame_id, x_range, y_range):
        if frame_id == 0:
            return {1: [10, 10, 0]}, [(1, 0, 0)]
        elif frame_id == 1:
            return {1: [9, 10, 0]}, [(1, 0, 0)]
        elif frame_id == 2:
            return {1: [8, 10, 0]}, [(1, 0, 0)]
        elif frame_id == 3:
            return {1: [14, 12, 0]}, [(1, 0, 0)]
        else:
            return {1: [10, 10, 0]}, [(1, 0, 0)]

    def add_agents(self, x, y, g):
        pass

    def get_map_size(self):
        return [50, 50]



"""deprecated"""

import numpy as np

from magent.model import BaseModel
from magent.c_lib import _LIB, as_int32_c_array, as_float_c_array


class RunawayPrey(BaseModel):
    def __init__(self, env, handle, away_handle, *args, **kwargs):
        BaseModel.__init__(self, env, handle)

        self.away_channel = env.get_channel(away_handle)
        self.attack_base, _ = env.get_view2attack(handle)
        self.move_back = 4

        print("attack base", self.attack_base, "away", self.away_channel)

    def infer_action(self, observations, *args, **kwargs):
        obs_buf = as_float_c_array(observations[0])
        hp_buf  = as_float_c_array(observations[1])
        n, height, width, n_channel = observations[0].shape
        buf = np.empty((n,), dtype=np.int32)
        act_buf = as_int32_c_array(buf)
        _LIB.runaway_infer_action(obs_buf, hp_buf, n, height, width, n_channel,
                                  self.attack_base, act_buf, self.away_channel, self.move_back)
        return buf



import os
import mxnet as mx

from magent.utility import has_gpu
from magent.model import BaseModel


class MXBaseModel(BaseModel):
    def __init__(self, env, handle, name, subclass_name):
        """init a model

        Parameters
        ----------
        env: magent.Environment
        handle: handle (ctypes.c_int32)
        name: str
        subclass_name: str
            name of subclass
        """
        BaseModel.__init__(self, env, handle)
        self.name = name
        self.subclass_name = subclass_name

    def _get_ctx(self):
        """return correct context , priority: gpu > cpu

        Returns
        -------
        ctx: mx.context
        """
        if has_gpu():
            return mx.gpu()
        else:
            return mx.cpu()

    def save(self, dir_name, epoch):
        """save model to dir

        Parameters
        ----------
        dir_name: str
            name of the directory
        epoch: int
        """
        if not os.path.exists(dir_name):
            os.mkdir(dir_name)
        dir_name = os.path.join(dir_name, self.name, )
        if not os.path.exists(dir_name):
            os.mkdir(dir_name)
        pre = os.path.join(dir_name, self.subclass_name)
        self.model.save_checkpoint(pre, epoch, save_optimizer_states=True)

    def load(self, dir_name, epoch=0, name=None):
        """save model to dir

        Parameters
        ----------
        dir_name: str
            name of the directory
        epoch: int
        """
        name = name or self.name
        dir_name = os.path.join(dir_name, name)
        pre = os.path.join(dir_name, self.subclass_name)
        _, arg_params, aux_params = mx.model.load_checkpoint(pre, epoch)
        self.model.set_params(arg_params, aux_params, force_init=True)



""" battle of two armies """

import magent


def get_config(map_size):
    gw = magent.gridworld
    cfg = gw.Config()

    cfg.set({"map_width": map_size, "map_height": map_size})
    cfg.set({"minimap_mode": True})
    cfg.set({"embedding_size": 10})

    small = cfg.register_agent_type(
        "small",
        {'width': 1, 'length': 1, 'hp': 10, 'speed': 2,
         'view_range': gw.CircleRange(6), 'attack_range': gw.CircleRange(1.5),
         'damage': 2, 'step_recover': 0.1,

         'step_reward': -0.005,  'kill_reward': 5, 'dead_penalty': -0.1, 'attack_penalty': -0.1,
         })

    g0 = cfg.add_group(small)
    g1 = cfg.add_group(small)

    a = gw.AgentSymbol(g0, index='any')
    b = gw.AgentSymbol(g1, index='any')

    # reward shaping to encourage attack
    cfg.add_reward_rule(gw.Event(a, 'attack', b), receiver=a, value=0.2)
    cfg.add_reward_rule(gw.Event(b, 'attack', a), receiver=b, value=0.2)

    return cfg



""" tigers eat deer to get health point and reward"""

import magent


def get_config(map_size):
    gw = magent.gridworld

    cfg = gw.Config()

    cfg.set({"map_width": map_size, "map_height": map_size})
    cfg.set({"embedding_size": 10})

    deer = cfg.register_agent_type(
        "deer",
        {'width': 1, 'length': 1, 'hp': 5, 'speed': 1,
         'view_range': gw.CircleRange(1), 'attack_range': gw.CircleRange(0),
         'damage': 0, 'step_recover': 0.2,
         'food_supply': 0, 'kill_supply': 8,
         })

    tiger = cfg.register_agent_type(
        "tiger",
        {'width': 1, 'length': 1, 'hp': 10, 'speed': 1,
         'view_range': gw.CircleRange(4), 'attack_range': gw.CircleRange(1),
         'damage': 3, 'step_recover': -0.5,
         'food_supply': 0, 'kill_supply': 0,
         'step_reward': 1, 'attack_penalty': -0.1,
         })

    deer_group  = cfg.add_group(deer)
    tiger_group = cfg.add_group(tiger)

    return cfg



""" base model classes"""

try:
    import thread
except ImportError:
    import _thread as thread

import multiprocessing
import multiprocessing.connection
import sys

import numpy as np

class BaseModel:
    def __init__(self, env, handle, *args, **kwargs):
        """ init

        Parameters
        ----------
        env: Environment
            env
        handle: GroupHandle
            handle of this group, handles are returned by env.get_handles()
        """
        pass

    def infer_action(self, raw_obs, ids, *args, **kwargs):
        """ infer action for a group of agents

        Parameters
        ----------
        raw_obs: tuple
            raw_obs is a tuple of (view, feature)
            view is a numpy array, its shape is n * view_width * view_height * n_channel
                                   it contains the spatial local observation for all the agents
            feature is a numpy array, its shape is n * feature_size
                                   it contains the non-spatial feature for all the agents
        ids: numpy array of int32
            the unique id of every agents
        args:
            additional custom args
        kwargs:
            additional custom args
        """
        pass

    def train(self, sample_buffer, **kwargs):
        """ feed new samples and train

        Parameters
        ----------
        sample_buffer: EpisodesBuffer
            a buffer contains transitions of agents

        Returns
        -------
        loss and estimated mean state value
        """
        return 0, 0    # loss, mean value

    def save(self, *args, **kwargs):
        """ save the model """
        pass

    def load(self, *args, **kwargs):
        """ load the model """
        pass


class NDArrayPackage:
    """wrapper for transferring numpy arrays by bytes"""
    def __init__(self, *args):
        if isinstance(args[0], np.ndarray):
            self.data = args
            self.info = [(x.shape, x.dtype) for x in args]
        else:
            self.data = None
            self.info = args[0]

        self.max_len = (1 << 30) / 4

    def send_to(self, conn, use_thread=False):
        assert self.data is not None

        def send_thread():
            for x in self.data:
                if np.prod(x.shape)  > self.max_len:
                    seg = int(self.max_len // np.prod(x.shape[1:]))
                    for pt in range(0, len(x), seg):
                        conn.send_bytes(x[pt:pt+seg])
                else:
                    conn.send_bytes(x)

        if use_thread:
            thread.start_new_thread(send_thread, ())
        else:
            send_thread()

    def recv_from(self, conn):
        bufs = []
        for info in self.info:
            buf = np.empty(shape=(int(np.prod(info[0])),), dtype=info[1])

            item_size = int(np.prod(info[0][1:]))
            if np.prod(info[0]) > self.max_len:
                seg = int(self.max_len // item_size)
                for pt in range(0, int(np.prod(info[0])), seg * item_size):
                    conn.recv_bytes_into(buf[pt:pt+seg * item_size])
            else:
               conn.recv_bytes_into(buf)
            bufs.append(buf.reshape(info[0]))
        return bufs


class ProcessingModel(BaseModel):
    """
    start a sub-processing to host a model,
    use pipe or socket for communication
    """
    def __init__(self, env, handle, name, port, sample_buffer_capacity=1000,
                 RLModel=None, **kwargs):
        """
        Parameters
        ----------
        env: environment
        handle: group handle
        name: str
            name of the model (be used when store model)
        port: int
            port of socket or suffix of pipe
        sample_buffer_capacity: int
            the maximum number of samples (s,r,a,s') to collect in a game round
        RLModel: BaseModel
            the RL algorithm class
        kwargs: dict
            arguments for RLModel
        """
        BaseModel.__init__(self, env, handle)

        assert RLModel is not None

        kwargs['env'] = env
        kwargs['handle'] = handle
        kwargs['name'] = name
        addr = 'magent-pipe-' + str(port)  # named pipe
        # addr = ('localhost', port) # socket
        proc = multiprocessing.Process(
            target=model_client,
            args=(addr, sample_buffer_capacity, RLModel, kwargs),
        )

        self.client_proc = proc
        proc.start()
        listener = multiprocessing.connection.Listener(addr)
        self.conn = listener.accept()

    def sample_step(self, rewards, alives, block=True):
        """record a step (should be followed by check_done)

        Parameters
        ----------
        block: bool
            if it is True, the function call will block
            if it is False, the caller must call check_done() afterward
                            to check/consume the return message
        """
        package = NDArrayPackage(rewards, alives)
        self.conn.send(["sample", package.info])
        package.send_to(self.conn)

        if block:
            self.check_done()

    def infer_action(self, raw_obs, ids, policy='e_greedy', eps=0, block=True):
        """ infer action

        Parameters
        ----------
        policy: str
            can be 'e_greedy' or 'greedy'
        eps: float
            used when policy is 'e_greedy'
        block: bool
            if it is True, the function call will block, and return actions
            if it is False, the function call won't block, the caller
                            must call fetch_action() to get actions

        Returns
        -------
        actions: numpy array (int32)
            see above
        """

        package = NDArrayPackage(raw_obs[0], raw_obs[1], ids)
        self.conn.send(["act", policy, eps, package.info])
        package.send_to(self.conn, use_thread=True)

        if block:
            info = self.conn.recv()
            return NDArrayPackage(info).recv_from(self.conn)[0]
        else:
            return None

    def fetch_action(self):
        """ fetch actions , fetch action after calling infer_action(block=False)

        Returns
        -------
        actions: numpy array (int32)
        """
        info = self.conn.recv()
        return NDArrayPackage(info).recv_from(self.conn)[0]

    def train(self, print_every=5000, block=True):
        """ train new data samples according to the model setting

        Parameters
        ----------
        print_every: int
            print training log info every print_every batches

        """
        self.conn.send(['train', print_every])

        if block:
            return self.fetch_train()

    def fetch_train(self):
        """ fetch result of train after calling train(block=False)

        Returns
        -------
        loss: float
            mean loss
        value: float
            mean state value
        """
        return self.conn.recv()

    def save(self, save_dir, epoch, block=True):
        """ save model

        Parameters
        ----------
        block: bool
            if it is True, the function call will block
            if it is False, the caller must call check_done() afterward
                            to check/consume the return message
        """

        self.conn.send(["save", save_dir, epoch])
        if block:
            self.check_done()

    def load(self, save_dir, epoch, name=None, block=True):
        """ load model

        Parameters
        ----------
        name: str
            name of the model (set when stored name is not the same as self.name)
        block: bool
            if it is True, the function call will block
            if it is False, the caller must call check_done() afterward
                            to check/consume the return message
        """
        self.conn.send(["load", save_dir, epoch, name])
        if block:
            self.check_done()

    def check_done(self):
        """ check return message of sub processing """
        assert self.conn.recv() == 'done'

    def quit(self):
        """ quit """
        proc = self.client_proc
        self.client_proc = None
        self.conn.send(["quit"])
        proc.join()

    def __del__(self):
        """ quit in destruction """
        if self.client_proc is not None:
            quit()


def model_client(addr, sample_buffer_capacity, RLModel, model_args):
    """target function for sub-processing to host a model

    Parameters
    ----------
    addr: socket address
    sample_buffer_capacity: int
        the maximum number of samples (s,r,a,s') to collect in a game round
    RLModel: BaseModel
        the RL algorithm class
    args: dict
        arguments to RLModel
    """
    import magent.utility

    model = RLModel(**model_args)
    sample_buffer = magent.utility.EpisodesBuffer(capacity=sample_buffer_capacity)

    conn = multiprocessing.connection.Client(addr)

    while True:
        cmd = conn.recv()
        if cmd[0] == 'act':
            policy = cmd[1]
            eps = cmd[2]
            array_info = cmd[3]

            view, feature, ids = NDArrayPackage(array_info).recv_from(conn)
            obs = (view, feature)

            acts = model.infer_action(obs, ids, policy=policy, eps=eps)
            package = NDArrayPackage(acts)
            conn.send(package.info)
            package.send_to(conn)
        elif cmd[0] == 'train':
            print_every = cmd[1]
            total_loss, value = model.train(sample_buffer, print_every=print_every)
            sample_buffer = magent.utility.EpisodesBuffer(sample_buffer_capacity)
            conn.send((total_loss, value))
        elif cmd[0] == 'sample':
            array_info = cmd[1]
            rewards, alives = NDArrayPackage(array_info).recv_from(conn)
            sample_buffer.record_step(ids, obs, acts, rewards, alives)
            conn.send("done")
        elif cmd[0] == 'save':
            savedir = cmd[1]
            n_iter = cmd[2]
            model.save(savedir, n_iter)
            conn.send("done")
        elif cmd[0] == 'load':
            savedir = cmd[1]
            n_iter = cmd[2]
            name = cmd[3]
            model.load(savedir, n_iter, name)
            conn.send("done")
        elif cmd[0] == 'quit':
            break
        else:
            print("Error: Unknown command %s" % cmd[0])
            break



import os
import tensorflow as tf

from magent.model import BaseModel


class TFBaseModel(BaseModel):
    """base model for tensorflow model"""
    def __init__(self, env, handle, name, subclass_name):
        """init a model

        Parameters
        ----------
        env: magent.Environment
        handle: handle (ctypes.c_int32)
        name: str
        subclass_name: str
            name of subclass
        """
        BaseModel.__init__(self, env, handle)
        self.name = name
        self.subclass_name = subclass_name

    def save(self, dir_name, epoch):
        """save model to dir

        Parameters
        ----------
        dir_name: str
            name of the directory
        epoch: int
        """
        if not os.path.exists(dir_name):
            os.mkdir(dir_name)
        dir_name = os.path.join(dir_name, self.name)
        if not os.path.exists(dir_name):
            os.mkdir(dir_name)
        model_vars = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, self.name)
        saver = tf.train.Saver(model_vars)
        saver.save(self.sess, os.path.join(dir_name, (self.subclass_name + "_%d") % epoch))

    def load(self, dir_name, epoch=0, name=None):
        """save model to dir

        Parameters
        ----------
        dir_name: str
            name of the directory
        epoch: int
        """
        if name is None or name == self.name:  # the name of saved model is the same as ours
            dir_name = os.path.join(dir_name, self.name)
            model_vars = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, self.name)
            saver = tf.train.Saver(model_vars)
            saver.restore(self.sess, os.path.join(dir_name, (self.subclass_name + "_%d") % epoch))
        else:  # load a checkpoint with different name
            backup_graph = tf.get_default_graph()
            kv_dict = {}

            # load checkpoint from another saved graph
            with tf.Graph().as_default(), tf.Session() as sess:
                tf.train.import_meta_graph(os.path.join(dir_name, name, (self.subclass_name + "_%d") % epoch + ".meta"))
                dir_name = os.path.join(dir_name, name)
                model_vars = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, name)
                sess.run(tf.global_variables_initializer())
                saver = tf.train.Saver(model_vars)
                saver.restore(sess, os.path.join(dir_name, (self.subclass_name + "_%d") % epoch))
                for item in tf.global_variables():
                    kv_dict[item.name] = sess.run(item)

            # assign to now graph
            backup_graph.as_default()
            model_vars = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, self.name)
            for item in model_vars:
                old_name = item.name.replace(self.name, name)
                self.sess.run(tf.assign(item, kv_dict[old_name]))




"""plot a heatmap for tournament"""
import matplotlib.pyplot as plt
import numpy as np

def plot_heatmap(x, y, z):
    x, y = np.meshgrid(y, x)
    fig, ax = plt.subplots()
    im = ax.pcolormesh(x, y, z)
    fig.colorbar(im)

def smooth(data, alpha, beta=None):
    beta = beta or alpha
    for i in range(0, len(data)):
        for j in range(1, len(data[0])):
            data[i][j] = alpha * data[i][j-1] + (1-alpha) * data[i][j]

    for j in range(0, len(data[0])):
        for i in range(1, len(data)):
            data[i][j] = alpha * data[i-1][j] + (1-alpha) * data[i][j]

    return data


filename = "detail.log"

data = []
round2index = {}
ct = 0

with open(filename) as fin:
    for line in fin.readlines():
        item = line.split("\t")
        l = eval(item[0].split(":")[1])[2]
        r = eval(item[1].split(":")[1])[2]
        rate = eval(item[2].split(":")[1])
        num = eval(item[3].split(":")[1])

        for no in [l, r]:
            if no not in round2index:
                round2index[no] = ct
                ct += 1

        data.append([l, r, rate, num])

heat_data = [[0.5 for _ in range(ct)] for _ in range(ct)]

for line in data:
    l = round2index[line[0]]
    r = round2index[line[1]]
    rate = line[2]
    num = line[3]
    heat_data[l][r] = rate
    heat_data[r][l] = 1 - rate

heat_data = smooth(heat_data, 0.8)
heat_data = np.array(heat_data)
rounds = np.sort(np.array(round2index.keys()))

pick = 60
heat_data = heat_data[:pick,:pick]
rounds = rounds[:pick]

plot_heatmap(rounds, rounds, heat_data)

plt.show()



""" base class for environment """


class Environment:
    """see subclass for detailed comment"""
    def __init__(self):
        pass

    def reset(self):
        pass

    # ====== RUN ======
    def get_observation(self, handle):
        pass

    def set_action(self, handle, actions):
        pass

    def step(self):
        pass

    def render(self):
        pass

    def render_next_file(self):
        pass

    def get_reward(self, handle):
        pass

    # ====== INFO ======
    def get_num(self, handle):
        pass

    def get_action_space(self, handle):
        pass

    def get_view_space(self, handle):
        pass

    def get_feature_space(self, handle):
        pass




"""test baselines in double attack"""

from search import do_task

task = [
    {
        "name": "tiger",
        "type": "single-search",
        "prefix": "python examples/train_tiger.py --train --n_round 250",
        "arg_name": "--alg",
        "arg_value": ["dqn", "a2c", "drqn"]
    }
]

for item in task:
    do_task(item)
    print("%s done" % item['name'])

print("tiger all done")
print("plot curve: python scripts/plot_many.py tiger-rec.out reward")



import setuptools
from setuptools import Extension
from setuptools.command.install import install
from setuptools.command.develop import develop
from setuptools.command.install_scripts import install_scripts
from setuptools.command.build_ext import build_ext as build_ext_orig
from subprocess import check_call
from distutils.sysconfig import get_python_lib
import os
import platform
import pathlib
import atexit

with open("README.md", "r") as fh:
    long_description = fh.read()

class PostInstallCommand(install_scripts):
    def run(self):
        post_install_script(self.build_dir)
        install_scripts.run(self)
        
def post_install_script(s):
    current_dir = os.getcwd()
    site_p = pathlib.Path(s).parent
    print(site_p)
    for sub in site_p.iterdir():
        f = str(sub)
        if "bdist" in f:
            raw_build_dir = f+"/wheel/magent"
            break
    
    if raw_build_dir != "":
        os.chdir(str(raw_build_dir))
        check_call("bash build.sh".split())
        os.chdir(str(current_dir))
    else:
        print("BDIST unavailable for MAgent setup, cannot build.")


setuptools.setup(
    name="magent",
    version="0.1.3",
    author="PettingZoo Team",
    author_email="justinkterry@gmail.com",
    description="Multi-Agent Reinforcement Learning environments with very large numbers of agents",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/PettingZoo-Team/MAgent",
    keywords=["Reinforcement Learning", "game", "RL", "AI"],
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy>=1.18.0',
        'pygame>=2.0.0.dev6'
    ],
    python_requires='>=3.5',
    data_files = [("", ["LICENSE"])],
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    cmdclass = {
        'install_scripts': PostInstallCommand,
    },
)






"""dynamic plot class"""
import matplotlib.pyplot as plt


class DynamicPlot:
    def __init__(self, n):
        self.x_data  = []
        self.y_datas = []
        self.lines   = []

        plt.show()
        axes = plt.gca()

        for i in range(n):
            self.y_datas.append([])
            line, = axes.plot(self.x_data, self.y_datas[-1])
            self.lines.append(line)

        self.axes = axes

    def add_point(self, x, ys):
        self.x_data.append(x)

        for i in range(len(ys)):
            self.y_datas[i].append(ys[i])

    def redraw(self):
        for i in range(len(self.lines)):
            self.lines[i].set_xdata(self.x_data)
            self.lines[i].set_ydata(self.y_datas[i])

        self.axes.autoscale(True)
        self.axes.relim()
        x_left, x_right = self.axes.get_xlim()
        y_left, y_right = self.axes.get_ylim()
        self.axes.set_xlim(x_left, (int(x_right) / 100 + 1) * 100)
        self.axes.set_ylim(0, y_right * 1.2)
        plt.draw()
        plt.pause(1e-15)

    def save(self, filename):
        plt.savefig(filename)




import time

import numpy as np
import mxnet as mx

from .base import MXBaseModel
from ..common import ReplayBuffer
from ...utility import has_gpu


class DeepQNetwork(MXBaseModel):
    def __init__(self, env, handle, name,
                 batch_size=64, learning_rate=1e-4, reward_decay=0.99,
                 train_freq=1, target_update=2000, memory_size=2 ** 20, eval_obs=None,
                 use_dueling=True, use_double=True, infer_batch_size=8192,
                 custom_view_space=None, custom_feature_space=None, num_gpu=1):
        """init a model

        Parameters
        ----------
        env: Environment
            environment
        handle: Handle (ctypes.c_int32)
            handle of this group, can be got by env.get_handles
        name: str
            name of this model
        learning_rate: float
        batch_size: int
        reward_decay: float
            reward_decay in TD
        train_freq: int
            mean training times of a sample
        target_update: int
            target will update every target_update batches
        memory_size: int
            weight of entropy loss in total loss
        eval_obs: numpy array
            evaluation set of observation
        use_dueling: bool
            whether use dueling q network
        use_double: bool
            whether use double q network
        num_gpu: int
            number of gpu
        infer_batch_size: int
            batch size while inferring actions
        custom_feature_space: tuple
            customized feature space
        custom_view_space: tuple
            customized feature space
        """
        MXBaseModel.__init__(self, env, handle, name, "mxdqn")
        # ======================== set config  ========================
        self.env = env
        self.handle = handle
        self.view_space = custom_view_space or env.get_view_space(handle)
        self.feature_space = custom_feature_space or env.get_feature_space(handle)
        self.num_actions  = env.get_action_space(handle)[0]

        self.batch_size    = batch_size
        self.infer_batch_size = infer_batch_size
        self.learning_rate = learning_rate
        self.train_freq    = train_freq      # train time of every sample (s,a,r,s')
        self.target_update = target_update   # update frequency of target network
        self.eval_obs      = eval_obs
        self.num_gpu       = num_gpu

        self.use_dueling  = use_dueling
        self.use_double   = use_double

        self.train_ct = 0

        # ======================= build network =======================
        self.ctx = self._get_ctx()
        if self.num_gpu > 1 and self.ctx == mx.gpu():
            self.ctx = []
            for i in range(self.num_gpu):
                self.ctx.append(mx.gpu(i))

        self.input_view = mx.sym.var("input_view")
        self.input_feature = mx.sym.var("input_feature")
        self.mask   = mx.sym.var("mask")
        self.action = mx.sym.var("action")
        self.target = mx.sym.var("target")

        self.qvalues = self._create_network(self.input_view, self.input_feature)
        self.gamma = reward_decay
        self.action_onehot = mx.sym.one_hot(self.action, depth=self.num_actions)
        td_error = mx.sym.square(self.target - mx.sym.sum(self.qvalues * self.action_onehot, axis=1))
        self.loss = mx.sym.sum(td_error * self.mask) / mx.sym.sum(self.mask)
        self.loss = mx.sym.MakeLoss(data=self.loss)

        self.out_qvalues = mx.sym.BlockGrad(self.qvalues)
        self.output = mx.sym.Group([self.out_qvalues, self.loss])

        self.model = mx.mod.Module(self.output,
                                   data_names=['input_view', 'input_feature'],
                                   label_names=['action', 'target', 'mask'], context=self.ctx)

        self.target_model = mx.mod.Module(self.qvalues,
                                          data_names=['input_view', 'input_feature'],
                                          label_names=[], context=self.ctx)

        # bind (set initial batch size)
        self.bind_size = batch_size
        self.model.bind(data_shapes=[('input_view', (batch_size,) + self.view_space),
                                     ('input_feature', (batch_size,) + self.feature_space)],
                        label_shapes=[('action', (batch_size,)),
                                      ('target', (batch_size,)),
                                      ('mask', (batch_size,))])
        self.target_model.bind(data_shapes=[('input_view', (batch_size,) + self.view_space),
                                            ('input_feature', (batch_size,) + self.feature_space)])

        # init params
        self.model.init_params(initializer=mx.init.Xavier())
        self.model.init_optimizer(optimizer='adam', optimizer_params={
            'learning_rate': learning_rate,
            'clip_gradient': 10.0})

        self._copy_network(self.target_model, self.model)

        # init replay buffers
        self.replay_buf_len = 0
        self.memory_size = memory_size
        self.replay_buf_view     = ReplayBuffer(shape=(memory_size,) + self.view_space)
        self.replay_buf_feature  = ReplayBuffer(shape=(memory_size,) + self.feature_space)
        self.replay_buf_action   = ReplayBuffer(shape=(memory_size,), dtype=np.int32)
        self.replay_buf_reward   = ReplayBuffer(shape=(memory_size,))
        self.replay_buf_terminal = ReplayBuffer(shape=(memory_size,), dtype=np.bool)
        self.replay_buf_mask     = ReplayBuffer(shape=(memory_size,))
        # if mask[i] == 0, then the item is used for padding, not for training

        # print("parameters", self.model.get_params())
        # mx.viz.plot_network(self.loss).view()

    def _create_network(self, input_view, input_feature, use_conv=True):
        """define computation graph of network

        Parameters
        ----------
        input_view: mx.symbol
        input_feature: mx.symbol
            the input tensor
        """
        kernel_num = [32, 32]
        hidden_size = [256]

        if use_conv:
            input_view = mx.sym.transpose(data=input_view, axes=[0, 3, 1, 2])
            h_conv1 = mx.sym.Convolution(data=input_view, kernel=(3, 3),
                                         num_filter=kernel_num[0], layout="NCHW")
            h_conv1 = mx.sym.Activation(data=h_conv1, act_type="relu")
            h_conv2 = mx.sym.Convolution(data=h_conv1, kernel=(3, 3),
                                         num_filter=kernel_num[1], layout="NCHW")
            h_conv2 = mx.sym.Activation(data=h_conv2, act_type="relu")
        else:
            input_view = mx.sym.flatten(data=input_view)
            h_conv2 = mx.sym.FullyConnected(input_view, num_hidden=hidden_size[0])
            h_conv2 = mx.sym.Activation(data=h_conv2, act_type="relu")

        flatten_view = mx.sym.flatten(data=h_conv2)
        h_view = mx.sym.FullyConnected(data=flatten_view, num_hidden=hidden_size[0])
        h_view = mx.sym.Activation(data=h_view, act_type="relu")

        h_emb = mx.sym.FullyConnected(data=input_feature, num_hidden=hidden_size[0])
        h_emb = mx.sym.Activation(data=h_emb, act_type="relu")

        dense = mx.sym.concat(h_view, h_emb)

        if self.use_dueling:
            # state value
            value = mx.sym.FullyConnected(data=dense, num_hidden=1)
            advantage = mx.sym.FullyConnected(data=dense, num_hidden=self.num_actions)

            mean = mx.sym.mean(advantage, axis=1, keepdims=True)
            advantage = mx.sym.broadcast_sub(advantage, mean)
            qvalues = mx.sym.broadcast_add(advantage, value)
        else:
            qvalues = mx.sym.FullyConnected(data=dense, num_hidden=self.num_actions)

        return qvalues

    def infer_action(self, raw_obs, ids, policy="e_greedy", eps=0):
        """infer action for a batch of agents

        Parameters
        ----------
        raw_obs: tuple(numpy array, numpy array)
            raw observation of agents tuple(views, features)
        ids: numpy array
            ids of agents
        policy: str
            can be eps-greedy or greedy
        eps: float
            used when policy is eps-greedy

        Returns
        -------
        acts: numpy array of int32
            actions for agents
        """
        view, feature = raw_obs[0], raw_obs[1]

        if policy == 'e_greedy':
            eps = eps
        elif policy == 'greedy':
            eps = 0

        n = len(view)
        if n < self.num_gpu:
            view = np.tile(view, (self.num_gpu, 1, 1, 1))
            feature = np.tile(feature, (self.num_gpu, 1))

        batch_size = min(len(view), self.infer_batch_size)
        self._reset_bind_size(batch_size)
        best_actions = []
        infer_iter = mx.io.NDArrayIter(data=[view, feature], batch_size=batch_size)
        for batch in infer_iter:
            self.model.forward(batch, is_train=False)
            qvalue_batch = self.model.get_outputs()[0]
            batch_action = mx.nd.argmax(qvalue_batch, axis=1)
            best_actions.append(batch_action)
        best_actions = np.array([x.asnumpy() for x in best_actions]).flatten()
        best_actions = best_actions[:n]

        random = np.random.randint(self.num_actions, size=(n,))
        cond = np.random.uniform(0, 1, size=(n,)) < eps
        ret = np.where(cond, random, best_actions)

        return ret.astype(np.int32)

    def _calc_target(self, next_view, next_feature, rewards, terminal):
        """calculate target value"""
        n = len(rewards)

        data_batch = mx.io.DataBatch(data=[mx.nd.array(next_view), mx.nd.array(next_feature)])
        self._reset_bind_size(n)
        if self.use_double:
            self.target_model.forward(data_batch, is_train=False)
            self.model.forward(data_batch, is_train=False)
            t_qvalues = self.target_model.get_outputs()[0].asnumpy()
            qvalues   = self.model.get_outputs()[0].asnumpy()
            next_value = t_qvalues[np.arange(n), np.argmax(qvalues, axis=1)]
        else:
            self.target_model.forward(data_batch, is_train=False)
            t_qvalues = self.target_model.get_outputs()[0].asnumpy()
            next_value = np.max(t_qvalues, axis=1)

        target = np.where(terminal, rewards, rewards + self.gamma * next_value)

        return target

    def _add_to_replay_buffer(self, sample_buffer):
        """add samples in sample_buffer to replay buffer"""
        n = 0
        for episode in sample_buffer.episodes():
            v, f, a, r = episode.views, episode.features, episode.actions, episode.rewards

            m = len(r)

            mask = np.ones((m,))
            terminal = np.zeros((m,), dtype=np.bool)
            if episode.terminal:
                terminal[-1] = True
            else:
                mask[-1] = 0

            self.replay_buf_view.put(v)
            self.replay_buf_feature.put(f)
            self.replay_buf_action.put(a)
            self.replay_buf_reward.put(r)
            self.replay_buf_terminal.put(terminal)
            self.replay_buf_mask.put(mask)

            n += m

        self.replay_buf_len = min(self.memory_size, self.replay_buf_len + n)
        return n

    def train(self, sample_buffer, print_every=1000):
        """ add new samples in sample_buffer to replay buffer and train

        Parameters
        ----------
        sample_buffer: magent.utility.EpisodesBuffer
            buffer contains samples
        print_every: int
            print log every print_every batches

        Returns
        -------
        loss: float
            bellman residual loss
        value: float
            estimated state value
        """
        add_num = self._add_to_replay_buffer(sample_buffer)
        batch_size = self.batch_size
        total_loss = 0

        n_batches = int(self.train_freq * add_num / batch_size)
        if n_batches == 0:
            return 0, 0

        print("batch number: %d  add: %d  replay_len: %d/%d" %
              (n_batches, add_num, self.replay_buf_len, self.memory_size))

        start_time = time.time()
        ct = 0
        for i in range(n_batches):
            # fetch a batch
            index = np.random.choice(self.replay_buf_len - 1, batch_size)

            batch_view     = self.replay_buf_view.get(index)
            batch_feature  = self.replay_buf_feature.get(index)
            batch_action   = self.replay_buf_action.get(index)
            batch_reward   = self.replay_buf_reward.get(index)
            batch_terminal = self.replay_buf_terminal.get(index)
            batch_mask     = self.replay_buf_mask.get(index)

            batch_next_view    = self.replay_buf_view.get(index+1)
            batch_next_feature = self.replay_buf_feature.get(index+1)

            batch_target = self._calc_target(batch_next_view, batch_next_feature,
                                             batch_reward, batch_terminal)

            self._reset_bind_size(batch_size)
            batch = mx.io.DataBatch(data=[mx.nd.array(batch_view),
                                          mx.nd.array(batch_feature)],
                                    label=[mx.nd.array(batch_action),
                                           mx.nd.array(batch_target),
                                           mx.nd.array(batch_mask)])
            self.model.forward(batch, is_train=True)
            self.model.backward()
            self.model.update()
            loss = np.mean(self.model.get_outputs()[1].asnumpy())
            total_loss += loss

            if ct % self.target_update == 0:
                self._copy_network(self.target_model, self.model)

            if ct % print_every == 0:
                print("batch %5d,  loss %.6f, eval %.6f" % (ct, loss, self._eval(batch_target)))
            ct += 1
            self.train_ct += 1

        total_time = time.time() - start_time
        step_average = total_time / max(1.0, (ct / 1000.0))
        print("batches: %d,  total time: %.2f,  1k average: %.2f" % (ct, total_time, step_average))

        return total_loss / ct if ct != 0 else 0, self._eval(batch_target)

    def _reset_bind_size(self, new_size):
        """reset batch size"""
        if self.bind_size == new_size:
            return
        else:
            self.bind_size = new_size
            def _reshape(model, is_target):
                data_shapes = [('input_view',    (new_size,) + self.view_space),
                               ('input_feature', (new_size,) + self.feature_space)]
                label_shapes = [('action',        (new_size,)),
                                ('target',        (new_size,)),
                                ('mask',          (new_size,))]
                if is_target:
                    label_shapes = None
                model.reshape(data_shapes=data_shapes, label_shapes=label_shapes)
            _reshape(self.model, False)
            _reshape(self.target_model, True)

    def _copy_network(self, dest, source):
        """copy to target network"""
        arg_params, aux_params = source.get_params()
        dest.set_params(arg_params, aux_params)

    def _eval(self, target):
        """evaluate estimated q value"""
        if self.eval_obs is None:
            return np.mean(target)
        else:
            self._reset_bind_size(len(self.eval_obs[0]))
            with self.ctx:
                batch = mx.io.DataBatch(data=[mx.nd.array(self.eval_obs[0]),
                                              mx.nd.array(self.eval_obs[1])])
                self.model.forward(batch, is_train=False)
                return np.mean(self.model.get_outputs()[0].asnumpy())

    def get_info(self):
        return "mx dqn train_time: %d" % (self.train_ct)


"""gather agent, rush to food according to minimap"""

import numpy as np

from magent.model import BaseModel
from magent.c_lib import _LIB, as_int32_c_array, as_float_c_array


class RushGatherer(BaseModel):
    def __init__(self, env, handle, *args, **kwargs):
        BaseModel.__init__(self, env, handle)

        self.env = env
        self.handle = handle
        self.n_action = env.get_action_space(handle)
        self.view_size = env.get_view_space(handle)
        self.attack_base, self.view2attack = env.get_view2attack(handle)

    def infer_action(self, states, *args, **kwargs):
        obs_buf = as_float_c_array(states[0])
        hp_buf  = as_float_c_array(states[1])
        n, height, width, n_channel = states[0].shape
        buf = np.empty((n,), dtype=np.int32)
        act_buf = as_int32_c_array(buf)
        attack_base = self.attack_base
        
        view2attack_buf = as_int32_c_array(self.view2attack)

        _LIB.gather_infer_action(obs_buf, hp_buf, n, height, width, n_channel,
                                 act_buf, attack_base, view2attack_buf)
        return buf



""" advantage actor critic """
import os

import numpy as np
import tensorflow as tf

from .base import TFBaseModel


class AdvantageActorCritic(TFBaseModel):
    def __init__(self, env, handle, name, learning_rate=1e-3,
                 batch_size=64, reward_decay=0.99, eval_obs=None,
                 train_freq=1, value_coef=0.1, ent_coef=0.08, use_comm=False,
                 custom_view_space=None, custom_feature_space=None):
        """init a model

        Parameters
        ----------
        env: Environment
            environment
        handle: Handle (ctypes.c_int32)
            handle of this group, can be got by env.get_handles
        name: str
            name of this model
        learning_rate: float
        batch_size: int
        reward_decay: float
            reward_decay in TD
        eval_obs: numpy array
            evaluation set of observation
        train_freq: int
            mean training times of a sample
        ent_coef: float
            weight of entropy loss in total loss
        value_coef: float
            weight of value loss in total loss
        use_comm: bool
            whether use CommNet
        custom_feature_space: tuple
            customized feature space
        custom_view_space: tuple
            customized feature space
        """
        TFBaseModel.__init__(self, env, handle, name, "tfa2c")
        # ======================== set config  ========================
        self.env = env
        self.handle = handle
        self.name = name
        self.view_space = custom_view_space or env.get_view_space(handle)
        self.feature_space = custom_feature_space or env.get_feature_space(handle)
        self.num_actions  = env.get_action_space(handle)[0]
        self.reward_decay = reward_decay

        self.batch_size   = batch_size
        self.learning_rate= learning_rate
        self.train_freq   = train_freq   # train time of every sample (s,a,r,s')

        self.value_coef = value_coef     # coefficient of value in the total loss
        self.ent_coef = ent_coef         # coefficient of entropy in the total loss

        self.train_ct = 0
        self.use_comm = use_comm

        # ======================= build network =======================
        with tf.name_scope(self.name):
            self._create_network(self.view_space, self.feature_space)

        # init tensorflow session
        config = tf.ConfigProto(allow_soft_placement=True, log_device_placement=False)
        config.gpu_options.allow_growth = True
        self.sess = tf.Session(config=config)
        self.sess.run(tf.global_variables_initializer())

        # init training buffers
        self.view_buf = np.empty((1,) + self.view_space)
        self.feature_buf = np.empty((1,) + self.feature_space)
        self.action_buf = np.empty(1, dtype=np.int32)
        self.reward_buf = np.empty(1, dtype=np.float32)

    def _commnet_block(self, n, hidden, skip, name, hidden_size):
        """a block of CommNet

        Parameters
        ----------
        n: int
            number of agent
        hidden: tf.tensor
            hidden layer input
        skip: tf.tensor
            skip connection
        name: str
        hidden_size: int
        """
        mask = (tf.ones((n, n)) - tf.eye(n))
        mask *= tf.where(n > 1, 1.0 / (tf.cast(n, tf.float32) - 1.0), 0)

        C = tf.get_variable(name + "_C", shape=(hidden_size, hidden_size))
        H = tf.get_variable(name + "_H", shape=(hidden_size, hidden_size))

        message = tf.matmul(mask, hidden)

        return tf.tanh(tf.matmul(message, C) + tf.matmul(hidden, H) + skip)

    def _commnet(self, n, dense, hidden_size, n_step=2):
        """ CommNet Learning Multiagent Communication with Backpropagation by S. Sukhbaatar et al. NIPS 2016

        Parameters
        ----------
        n: int
            number of agent
        hidden_size: int
        n_step: int
            communication step

        Returns
        -------
        h: tf.tensor
            hidden units after CommNet
        """
        skip = dense

        h = dense
        for i in range(n_step):
            h = self._commnet_block(n, h, skip, "step_%d" % i, hidden_size)

        return h

    def _create_network(self, view_space, feature_space):
        """define computation graph of network

        Parameters
        ----------
        view_space: tuple
        feature_space: tuple
            the input shape
        """
        # input
        input_view    = tf.placeholder(tf.float32, (None,) + view_space)
        input_feature = tf.placeholder(tf.float32, (None,) + feature_space)
        action = tf.placeholder(tf.int32, [None])
        reward = tf.placeholder(tf.float32, [None])
        num_agent = tf.placeholder(tf.int32, [])

        kernel_num = [32, 32]
        hidden_size = [256]

        # fully connected
        flatten_view = tf.reshape(input_view, [-1, np.prod([v.value for v in input_view.shape[1:]])])
        h_view = tf.layers.dense(flatten_view, units=hidden_size[0], activation=tf.nn.relu)

        h_emb = tf.layers.dense(input_feature,  units=hidden_size[0], activation=tf.nn.relu)

        dense = tf.concat([h_view, h_emb], axis=1)
        dense = tf.layers.dense(dense, units=hidden_size[0] * 2, activation=tf.nn.relu)

        if self.use_comm:
            dense = self._commnet(num_agent, dense, dense.shape[-1].value)

        policy = tf.layers.dense(dense, units=self.num_actions, activation=tf.nn.softmax)
        policy = tf.clip_by_value(policy, 1e-10, 1-1e-10)
        value = tf.layers.dense(dense, units=1)
        value = tf.reshape(value, (-1,))
        advantage = tf.stop_gradient(reward - value)

        action_mask = tf.one_hot(action, self.num_actions)

        log_policy = tf.log(policy + 1e-6)
        log_prob = tf.reduce_sum(log_policy * action_mask, axis=1)
        pg_loss = -tf.reduce_mean(advantage * log_prob)
        vf_loss = self.value_coef * tf.reduce_mean(tf.square(reward - value))
        neg_entropy = self.ent_coef * tf.reduce_mean(tf.reduce_sum(policy * log_policy, axis=1))
        total_loss = pg_loss + vf_loss + neg_entropy

        # train op (clip gradient)
        optimizer = tf.train.AdamOptimizer(learning_rate=self.learning_rate)
        gradients, variables = zip(*optimizer.compute_gradients(total_loss))
        gradients, _ = tf.clip_by_global_norm(gradients, 5.0)
        self.train_op = optimizer.apply_gradients(zip(gradients, variables))

        train_op = tf.train.AdamOptimizer(learning_rate=self.learning_rate).minimize(total_loss)

        self.input_view = input_view
        self.input_feature = input_feature
        self.action     = action
        self.reward     = reward
        self.num_agent  = num_agent

        self.policy, self.value = policy, value
        self.train_op = train_op
        self.pg_loss, self.vf_loss, self.reg_loss = pg_loss, vf_loss, neg_entropy
        self.total_loss = total_loss

    def infer_action(self, raw_obs, ids, *args, **kwargs):
        """infer action for a batch of agents

        Parameters
        ----------
        raw_obs: tuple(numpy array, numpy array)
            raw observation of agents tuple(views, features)
        ids: numpy array
            ids of agents

        Returns
        -------
        acts: numpy array of int32
            actions for agents
        """
        view, feature = raw_obs[0], raw_obs[1]
        n = len(view)

        policy = self.sess.run(self.policy, {self.input_view: view,
                                             self.input_feature: feature,
                                             self.num_agent: n})
        actions = np.arange(self.num_actions)

        ret = np.empty(n, dtype=np.int32)
        for i in range(n):
            ret[i] = np.random.choice(actions, p=policy[i])

        return ret

    def train(self, sample_buffer, print_every=1000):
        """feed new data sample and train

        Parameters
        ----------
        sample_buffer: magent.utility.EpisodesBuffer
            buffer contains samples

        Returns
        -------
        loss: list
            policy gradient loss, critic loss, entropy loss
        value: float
            estimated state value
        """
        # calc buffer size
        n = 0
        for episode in sample_buffer.episodes():
            n += len(episode.rewards)

        # resize to the new size
        self.view_buf.resize((n,) + self.view_space)
        self.feature_buf.resize((n,) + self.feature_space)
        self.action_buf.resize(n)
        self.reward_buf.resize(n)
        view, feature  = self.view_buf, self.feature_buf
        action, reward = self.action_buf, self.reward_buf

        ct = 0
        gamma = self.reward_decay
        # collect episodes from multiple separate buffers to a continuous buffer
        for episode in sample_buffer.episodes():
            v, f, a, r = episode.views, episode.features, episode.actions, episode.rewards
            m = len(episode.rewards)

            r = np.array(r)
            keep = self.sess.run(self.value, feed_dict={
                self.input_view: [v[-1]],
                self.input_feature: [f[-1]],
                self.num_agent: 1
            })[0]
            for i in reversed(range(m)):
                keep = keep * gamma + r[i]
                r[i] = keep

            view[ct:ct+m] = v
            feature[ct:ct+m] = f
            action[ct:ct+m]  = a
            reward[ct:ct+m] = r
            ct += m

        assert n == ct

        # train
        _, pg_loss, vf_loss, ent_loss, state_value = self.sess.run(
            [self.train_op, self.pg_loss, self.vf_loss, self.reg_loss, self.value], feed_dict={
                self.input_view:    view,
                self.input_feature: feature,
                self.action:        action,
                self.reward:        reward,
                self.num_agent:     len(reward)
        })
        print("sample", n, pg_loss, vf_loss, ent_loss)

        return [pg_loss, vf_loss, ent_loss], np.mean(state_value)

    def get_info(self):
        """get information of the model

        Returns
        -------
        info: string
        """
        return "a2c train_time: %d" % (self.train_ct)
